#!/usr/bin/env python
"""
Utilities related to file handling
"""


import io
import os
import glob
import stat
import subprocess
import time
import zlib

from Utils.Utilities import decodeBytesToUnicode


def findFiles(path, pat):
    """
    Find files within given path and matching given pattern.
    :param path: starting directory path (string)
    :param pat: match pattern (string), e.g. *.py or name of the file
    :return: matched file names
    """
    files = []
    for idir, _, _ in os.walk(path):
        files.extend(glob.glob(os.path.join(idir, pat)))
    return files


def tarMode(tfile, opMode):
    """
    Extract proper mode of operation for given tar file. For instance,
    if op='r' and tfile name is file.tar.gz we should get 'r:gz',
    while if tfile name is file.tar.bz2 we should get 'r':bz2', while
    if tfile name is file.tar we should get 'r', etc.
    :param opMode: mode of operation (string), e.g. 'r', or 'w'
    :param tfile: sandbox tar file name (string)
    :return: mode of operation
    """
    ext = tfile.split(".")[-1]
    if ext == 'tar':
        return opMode
    mode = opMode + ':' + ext
    return mode


def calculateChecksums(filename):
    """
    _calculateChecksums_

    Get the adler32 and crc32 checksums of a file. Return None on error

    Process line by line and adjust for known signed vs. unsigned issues
      http://docs.python.org/library/zlib.html

    The cksum UNIX command line tool implements a CRC32 checksum that is
    different than any of the python algorithms, therefore open cksum
    in a subprocess and feed it the same chunks of data that are used
    to calculate the adler32 checksum.

    """
    adler32Checksum = 1  # adler32 of an empty string
    cksumProcess = subprocess.Popen("cksum", stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    # the lambda basically creates an iterator function with zero
    # arguments that steps through the file in 4096 byte chunks
    with open(filename, 'rb') as f:
        for chunk in iter((lambda: f.read(4096)), b''):
            adler32Checksum = zlib.adler32(chunk, adler32Checksum)
            cksumProcess.stdin.write(chunk)

    cksumProcess.stdin.close()
    cksumProcess.wait()

    cksumStdout = cksumProcess.stdout.read().split()
    cksumProcess.stdout.close()

    # consistency check on the cksum output
    filesize = os.stat(filename)[stat.ST_SIZE]
    if len(cksumStdout) != 2 or int(cksumStdout[1]) != filesize:
        raise RuntimeError("Something went wrong with the cksum calculation !")

    cksumStdout[0] = decodeBytesToUnicode(cksumStdout[0])
    return (format(adler32Checksum & 0xffffffff, '08x'), cksumStdout[0])


def tail(filename, nLines=20):
    """
    _tail_

    A version of tail
    Adapted from code on http://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-with-python-similar-to-tail
    """
    assert nLines >= 0
    pos, lines = nLines + 1, []

    # make sure only valid utf8 encoded chars will be passed along
    with io.open(filename, 'r', encoding='utf8', errors='ignore') as f:
        while len(lines) <= nLines:
            try:
                f.seek(-pos, 2)
            except IOError:
                f.seek(0)
                break
            finally:
                lines = list(f)
            pos *= 2

    text = "".join(lines[-nLines:])

    return text


def getFileInfo(filename):
    """
    _getFileInfo_

    Return file info in a friendly format
    """

    filestats = os.stat(filename)

    fileInfo = {'Name': filename,
                'Size': filestats[stat.ST_SIZE],
                'LastModification': time.strftime("%m/%d/%Y %I:%M:%S %p", time.localtime(filestats[stat.ST_MTIME])),
                'LastAccess': time.strftime("%m/%d/%Y %I:%M:%S %p", time.localtime(filestats[stat.ST_ATIME]))}
    return fileInfo


def findMagicStr(filename, matchString):
    """
    _findMagicStr_

    Parse a log file looking for a pattern string
    """
    with io.open(filename, 'r', encoding='utf8', errors='ignore') as logfile:
        # TODO: can we avoid reading the whole file
        for line in logfile:
            if matchString in line:
                yield line


def getFullPath(name, envPath="PATH"):
    """
    :param name: file name
    :param envPath: any environment variable specified for path (PATH, PYTHONPATH, etc)
    :return: full path if it is under PATH env
    """
    for path in os.getenv(envPath).split(os.path.pathsep):
        fullPath = os.path.join(path, name)
        if os.path.exists(fullPath):
            return fullPath
    return None



"""
File       : MSPileupData.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileupData provides logic behind
data used and stored by MSPileup module.

The data flow should be done via list of objects, e.g.
[<pileup object 1>, <pileup object 2>, ..., <pileup object n>]
"""

# system modules
import re

# third party modules
from pymongo import IndexModel, errors

# WMCore modules
from Utils.Timers import gmtimeSeconds
from WMCore.Database.MongoDB import MongoDB
from WMCore.MicroService.MSPileup.DataStructs.MSPileupObj import MSPileupObj, schema
from WMCore.MicroService.MSPileup.MSPileupError import MSPileupUniqueConstrainError, \
    MSPileupDatabaseError, MSPileupNoKeyFoundError, MSPileupDuplicateDocumentError, \
    MSPileupSchemaError, MSPileupGenericError, MSPileupFractionError
from WMCore.MicroService.Tools.Common import getMSLogger


def stripKeys(docs, skeys=None):
    """
    Helper function to strip out keys from given dictionary.

    :param docs: either input dictionary or list of dictionaries
    :param skeys: list of of strip keys
    """
    if skeys is None:
        return docs
    if isinstance(docs, list):
        results = []
        for doc in docs:
            if doc:
                for key in skeys:
                    doc.pop(key, None)
                    results.append(doc)
        return results

    if docs and isinstance(docs, dict):
        for key in skeys:
            docs.pop(key, None)
    return docs


def getNewTimestamp(doc):
    """
    Given a pileup doc - or a subset of it - return a dictionary
    with a couple timestamp attributes that need to be updated.
    :param doc: a python dictionary representing the pileup information
    :return: a python dictionary to update the pileup object
    """
    subDoc = {'lastUpdateTime': gmtimeSeconds()}
    if "active" in doc and doc['active'] is True:
        subDoc['activatedOn'] = gmtimeSeconds()
    elif "active" in doc and doc['active'] is False:
        subDoc['deactivatedOn'] = gmtimeSeconds()
    return subDoc


def customDID(pname):
    """
    Helper function to create custom Name from given pileup one
    """
    pat = re.compile(".*-V[0-9]+$")
    if re.match(pat, pname):
        arr = pname.split("-")
        lastSuffix = arr[-1]
        ver = int(lastSuffix.replace('V', ''))
        suffix = f"{ver + 1}"
        pname = ''.join(pname.split('-V')[0]) + '-V' + suffix
    else:
        pname += "-V1"
    return pname


def addTransitionRecord(doc, userDN, logger):
    """
    Add a new transition record if the current container fraction differs from the
    last transition record container fraction. If they are the same, nothing needs to
    be done with the transition records. For more information please see
    https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5

    :param doc: MSPileup record
    :param userDN: user DN (string)
    :param logger: logger to use
    :return: nothing (document transition is updated in place)
    """
    fraction = doc['containerFraction']
    prevTranRecord = doc['transition'][-1]
    userDN = userDN or prevTranRecord['DN']
    if prevTranRecord['containerFraction'] != fraction:
        customName = customDID(prevTranRecord['customDID'])
        # preserve previous container fraction
        transitionRecord = {'containerFraction': prevTranRecord['containerFraction'],
                            'customDID': customName,
                            'updateTime': gmtimeSeconds(),
                            'DN': userDN}
        doc['transition'].append(transitionRecord)
        logger.info("Added transition record for pileup %s", doc['pileupName'])


class MSPileupData():
    """
    MSPileupData provides logic behind data used and stored by MSPileup module
    """

    def __init__(self, msConfig, **kwargs):
        """
        Constructor for MSPileupData
        """
        self.logger = getMSLogger(False)
        self.msConfig = msConfig
        self.msConfig.setdefault("mongoDBRetryCount", 3)
        self.msConfig.setdefault("mongoDBReplicaSet", None)
        self.msConfig.setdefault("mongoDBPort", None)
        self.msConfig.setdefault("mockMongoDB", False)

        # A full set of valid database connection parameters can be found at:
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html
        mongoIndex = IndexModel('pileupName', unique=True)
        collection = self.msConfig.get('mongoDBCollection', 'msPileupCollection')
        msOutDBConfig = {
            'database': self.msConfig['mongoDB'],
            'server': self.msConfig['mongoDBServer'],
            'replicaSet': self.msConfig['mongoDBReplicaSet'],
            'port': self.msConfig['mongoDBPort'],
            'username': self.msConfig['mongoDBUser'],
            'password': self.msConfig['mongoDBPassword'],
            'mockMongoDB': self.msConfig['mockMongoDB'],
            'connect': True,
            'directConnection': False,
            'logger': self.logger,
            'create': True,
            'collections': [(collection, mongoIndex)]}
        mongoDB = MongoDB(**msOutDBConfig)
        self.msDB = getattr(mongoDB, self.msConfig['mongoDB'])
        self.dbColl = self.msDB[collection]

    def createPileup(self, pdict, rseList, userDN=''):
        """
        Create and return pileup data from campaigns dictionary

        :param pdict: a dictionary representing MSPileup data
        :param rseList: a list of RSE names
        :param userDN: a string, user DN used in HTTP PUT request
        :return: list of MSPileupError or empty list
        """
        # first, create MSPileupObj which will be validated against its schema
        # and against the valid and existent Rucio RSEs
        try:
            obj = MSPileupObj(pdict, validRSEs=rseList)
            doc = obj.getPileupData()
        except Exception as exp:
            msg = f"Failed to create MSPileupObj, {exp}"
            self.logger.exception(msg)
            err = MSPileupSchemaError(pdict, msg)
            self.logger.error(err)
            return [err.error()]

        # add initial transtition record according to logic:
        # https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        transitionRecord = {'containerFraction': doc.get('containerFraction', 1.0),
                            'customDID': doc['pileupName'],
                            'updateTime': gmtimeSeconds(),
                            'DN': userDN}
        doc['transition'] = [transitionRecord]

        # insert document into underlying DB
        try:
            self.dbColl.insert_one(doc)
            self.logger.info("Pileup object '%s' (custom name '%s') successfully created", doc.get("pileupName"), doc.get('customName'))
        except errors.DuplicateKeyError:
            msg = f"Failed to insert: {doc}, it already exist in DB"
            self.logger.exception(msg)
            err = MSPileupDuplicateDocumentError(doc, msg)
            self.logger.error(err)
            return [err.error()]
        return []

    def updatePileup(self, doc, rseList=None, validate=True, userDN=''):
        """
        Update pileup data with provided input. The given doc should be either
        full pileup document or contains mandatory fields like pileupName
        and containerFraction in case of partial pileup data-placement.

        :param doc: a dictionary of pieleup data to be updated
        :param rseList: a list of valid RSE names
        :param validate: boolean defining whether the doc needs
                         to be validated or not
        :param userDN: a string, user DN used in HTTP PUT request
        :return: list of MSPileupError or empty list
        """
        self.logger.info("Updating pileup document %s", doc)
        rseList = rseList or []

        # check if either pileup
        if doc.get('pileupName', '') == '':
            err = MSPileupNoKeyFoundError(doc, 'pileup name is not provided')
            self.logger.error(err)
            return [err.error()]

        # check that spec should not allow customName without pileupName
        if doc.get('customName', '') != '' and 'pileupName' not in doc:
            err = MSPileupNoKeyFoundError(doc, 'custom name is not allowed in spec without pileup name')
            self.logger.error(err)
            return [err.error()]

        # it is replaced with partial pileup placement one, see
        # https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5

        # get document with pileup name
        pname = doc['pileupName']
        spec = {'pileupName': pname}

        # look-up pileup document in underlying DB
        self.logger.info("Fetching pileup document with the following spec: %s", spec)
        results = self.getPileup(spec)
        if not results:
            err = MSPileupNoKeyFoundError(spec, f'No document found for {spec} query')
            self.logger.error(err)
            return [err.error()]

        # we should have a single document corresponding to given pileup name
        if len(results) != 1:
            msg = f"Unique constraint violated for {pname}"
            err = MSPileupUniqueConstrainError(spec, msg)
            self.logger.error(err)
            return [err.error()]

        # Based on the document retrieved from the database, apply the user-related
        # updates and run this new data structure through the usual validation
        dbDoc = results[0]

        # get fraction value of provided document spec
        fraction = doc.get('containerFraction', 1.0)

        # perform check of input doc, is it partial pileup spec or not
        if set(doc.keys()) == set(["pileupName", "containerFraction"]):
            # check if given containerFraction is differ from existing document one
            if fraction == dbDoc.get('containerFraction', 1.0):
                msg = f"container fraction in provided spec {spec} is identical to one in MongoDB"
                err = MSPileupFractionError(spec, msg)
                self.logger.error(err)
                return [err.error()]
            self.logger.info("partial pileup spec: %s", doc)

        # Based on the document retrieved from the database, apply the user-related updates
        dbDoc.update(doc)

        # add transition record if doc contains new fraction
        transition = dbDoc.get('transition', [])
        if not transition:
            msg = f"To update pileup {pname} document we should already have transition record in place"
            err = MSPileupUniqueConstrainError(spec, msg)
            self.logger.error(err)
            return [err.error()]

        # update transition record if necessary
        addTransitionRecord(dbDoc, userDN, self.logger)
        output = self.updatePileupDocumentInDatabase(dbDoc, rseList)
        return output

    def updatePileupDocumentInDatabase(self, doc, rseList=None, validate=True):
        """
        Update given pileup document in database
        :param doc: pileup document (dictionary)
        :param rseList: a list of valid RSE names
        :param validate: boolean flag to validate document (default)
        :return: outcome of update, either empty list in case of success or list of error
        """
        spec = {'pileupName': doc['pileupName']}
        partialPileupSpec = False
        if set(doc.keys()) == set(["pileupName", "containerFraction"]):
            partialPileupSpec = True

        rseList = rseList or []
        if validate:
            try:
                # NOTE: MSPileupObj calls validate in its ctor with provided validRSEs
                # therefore, for partial pileup we take expectedRSEs for validation
                # while for generic update case we use rseList passed to this API as is
                if partialPileupSpec or not rseList:
                    rseList = doc['expectedRSEs']
                obj = MSPileupObj(doc, validRSEs=rseList)
                doc = obj.getPileupData()
            except Exception as exp:
                msg = f"Failed to update MSPileupObj, {exp}"
                self.logger.exception(msg)
                err = MSPileupSchemaError(doc, msg)
                self.logger.error(err)
                return [err.error()]

        # mandatory timestamp updates
        doc.update(getNewTimestamp(doc))
        # we do not need to create MSPileupObj and validate it since our doc comes directly from DB
        try:
            self.logger.info("Updating pileup document: spec=%s doc=%s", spec, doc)
            self.dbColl.update_one(spec, {"$set": doc})
            self.logger.info("Pileup object '%s' (custom name '%s') successfully updated", spec.get("pileupName"), spec.get('customName'))
        except Exception as exp:
            msg = f"Failed to insert: {doc}, error {exp}"
            self.logger.exception(msg)
            err = MSPileupDatabaseError(doc, msg, exp)
            self.logger.error(err)
            return [err.error()]
        return []

    def sanitizeQuery(self, spec, projection=None):
        """
        Perform MSPileup query validation

        :param spec: input MongoDB query (JSON spec)
        :param projection: MongoDB projection,
        see https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html
        :return: list, either empty one in case of no errors or list with error dictionary
        """
        # check that given spec is dictionary object
        if not isinstance(spec, dict):
            msg = f"Failed to sanitize MSPileup query, given spec {spec} is not JSON dictionary"
            err = MSPileupGenericError(spec, msg)
            return [err.error()]

        # check that keys of our spec dictionary belong to MSPileup schema
        docSchema = schema()
        for key in spec.keys():
            if key not in docSchema.keys():
                msg = f"Failed to sanitize MSPileup query, invalid key {key}"
                err = MSPileupSchemaError(spec, msg)
                self.logger.error(err)
                return [err.error()]
        return []

    def getPileup(self, spec, projection=None):
        """
        Fetch MSPileup data from persistent storage for a given spec (JSON query)

        :param spec: input MongoDB query (JSON spec)
        :param projection: MongoDB projection,
        see https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html
        :return: list of documents fetched from database
        """
        results = []

        # remove API key from the spec as it defines pileup API
        if isinstance(spec, dict) and 'API' in spec:
            spec.pop('API')

        self.logger.info("Getting pileup with spec: %s, projection: %s", spec, projection)
        err = self.sanitizeQuery(spec, projection)
        if len(err) != 0:
            self.logger.error("Get pileup query didn't pass the validation. Error %s", err)
            return results

        for doc in self.dbColl.find(spec, projection):
            doc = stripKeys(doc, ['_id'])
            results.append(doc)
        return results

    def deletePileup(self, spec):
        """
        Delete MSPileup data in persistent storage for given spec (JSON query)

        :param spec: MongoDB spec (JSON query)
        :return: list of MSPileupError or empty list
        """
        try:
            self.dbColl.delete_one(spec)
            self.logger.info("Pileup object '%s' (custom name '%s') successfully deleted", spec.get("pileupName"), spec.get('cusomName'))
        except Exception as exp:
            msg = f"Unable to delete with spec {spec}, error {exp}"
            self.logger.exception(msg)
            err = MSPileupDatabaseError(spec, msg)
            self.logger.error(err)
            return [err.error()]
        return []



""" wmc-httpd [-q] { -r | --restart } [-l LOG-FILE] CONFIG-FILE
        wmc-httpd [-q] { -v | --verify } [-l LOG-FILE] CONFIG-FILE
        wmc-httpd [-q] { -s | --status } CONFIG-FILE
        wmc-httpd [-q] { -k | --kill } CONFIG-FILE

Manages a web server application. Loads configuration and all views, starting
up an appropriately configured CherryPy instance. Views are loaded dynamically
and can be turned on/off via configuration file."""

from __future__ import print_function
from builtins import object
from future.utils import viewitems
from future import standard_library
standard_library.install_aliases()

import errno
import logging
import os
import os.path
import re
import signal
import socket
import sys
import _thread
import time
import traceback
from argparse import ArgumentParser
from io import BytesIO, StringIO
from glob import glob
from subprocess import Popen, PIPE
from pprint import pformat

import cherrypy
from cherrypy import Application
from cherrypy._cplogging import LogManager
from cherrypy.lib import profiler

### Tools is needed for CRABServer startup: it sets up the tools attributes
import WMCore.REST.Tools
from WMCore.Configuration import ConfigSection, loadConfigurationFile
from Utils.Utilities import lowerCmsHeaders
from Utils.PythonVersion import PY2

#: Terminal controls to switch to "OK" status message colour.
COLOR_OK = "\033[0;32m"

#: Terminal controls to switch to "warning" status message colour.
COLOR_WARN = "\033[0;31m"

#: Terminal controls to restore normal message colour.
COLOR_NORMAL = "\033[0;39m"


def sig_terminate(signum=None, frame=None):
    """Termination signal handler.

    CherryPy termination signal handling is broken, the handler does not
    take the right number of arguments.  This is our own fixed handler
    to terminate the web server gracefully; in theory it could be
    removed when CherryPy is fixed, but we attach other signals here
    and print a logging entry."""
    cherrypy.log("INFO: exiting server from termination signal %d" % signum, severity=logging.INFO)
    cherrypy.engine.exit()


def sig_reload(signum=None, frame=None):
    """SIGHUP handler to restart the server.

    This just adds some logging compared to the CherryPy signal handler."""
    cherrypy.log("INFO: restarting server from hang-up signal %d" % signum, severity=logging.INFO)
    cherrypy.engine.restart()


def sig_graceful(signum=None, frame=None):
    """SIGUSR1 handler to restart the server gracefully.

    This just adds some logging compared to the CherryPy signal handler."""
    cherrypy.log("INFO: restarting server gracefully from signal %d" % signum, severity=logging.INFO)
    cherrypy.engine.graceful()


class ProfiledApp(Application):
    """Wrapper CherryPy Application object which generates aggregated
    profiles for the component on each call. Note that there needs to
    be an instance of this for each mount point to be profiled."""

    def __init__(self, app, path):
        Application.__init__(self, app.root, app.script_name, app.config)
        self.profiler = profiler.ProfileAggregator(path)

    def __call__(self, env, handler):
        def gather(): return Application.__call__(self, env, handler)

        return self.profiler.run(gather)


class Logger(LogManager):
    """Custom logger to record information in format we prefer."""

    def __init__(self, *args, **kwargs):
        self.host = socket.gethostname()
        LogManager.__init__(self, *args, **kwargs)

    def access(self):
        """Record one client access."""
        request = cherrypy.request
        remote = request.remote
        response = cherrypy.response
        inheaders = lowerCmsHeaders(request.headers)
        outheaders = response.headers
        wfile = request.wsgi_environ.get('cherrypy.wfile', None)
        nout = (wfile and wfile.bytes_written) or outheaders.get('Content-Length', 0)
        if hasattr(request, 'start_time'):
            delta_time = (time.time() - request.start_time) * 1e6
        else:
            delta_time = 0
        msg = ('%(t)s %(H)s %(h)s "%(r)s" %(s)s'
               ' [data: %(i)s in %(b)s out %(T).0f us ]'
               ' [auth: %(AS)s "%(AU)s" "%(AC)s" ]'
               ' [ref: "%(f)s" "%(a)s" ]') % \
              {'t': self.time(),
               'H': self.host,
               'h': remote.name or remote.ip,
               'r': request.request_line,
               's': response.status,
               # request.rfile.rfile.bytes_read is a custom CMS web
               #  cherrypy patch not always available, hence the test
               'i': (getattr(request.rfile, 'rfile', None)
                     and getattr(request.rfile.rfile, "bytes_read", None)
                     and request.rfile.rfile.bytes_read) or "-",
               'b': nout or "-",
               'T': delta_time,
               'AS': inheaders.get("cms-auth-status", "-"),
               'AU': inheaders.get("cms-auth-cert", inheaders.get("cms-auth-host", "")),
               'AC': getattr(request.cookie.get("cms-auth", None), "value", ""),
               'f': inheaders.get("Referer", ""),
               'a': inheaders.get("User-Agent", "")}
        self.access_log.log(logging.INFO, msg)
        self.access_log.propagate = False  # to avoid duplicate records on the log
        self.error_log.propagate = False  # to avoid duplicate records on the log


class RESTMain(object):
    """Base class for the core CherryPy main application object.

    The :class:`~.RESTMain` implements basic functionality of a CherryPy-based
    server. Most users will want the fully functional :class:`~.RESTDaemon`
    instead; but in some cases such as tests and other single-shot jobs which
    don't require a daemon process this class is useful in its own right.

    The class implements the methods required to configure, but not run, a
    CherryPy server set up with an application configuration.

    The main application object takes the server configuration and state
    directory as parametres. It provides methods to create full CherryPy
    serer and configure the application based on configuration description."""

    def __init__(self, config, statedir):
        """Prepare the server.

        :arg config: server configuration
        :arg str statedir: server state directory."""
        self.config = config
        self.appname = config.main.application.lower()
        self.appconfig = config.section_(self.appname)
        self.srvconfig = config.section_("main")
        self.statedir = statedir
        self.hostname = socket.getfqdn().lower()
        self.silent = False
        self.extensions = {}
        self.views = {}

    def validate_config(self):
        """Check the server configuration has the required sections."""
        for key in ('admin', 'description', 'title'):
            if not hasattr(self.appconfig, key):
                raise RuntimeError("'%s' required in application configuration" % key)

    def setup_server(self):
        """Configure CherryPy server from application configuration.

        Traverses the server configuration portion and applies parameters
        known to be for CherryPy to the CherryPy server configuration.
        These are: engine, hooks, log, request, respose, server, tools,
        wsgi, checker.

        Also applies pseudo-parameters ``thread_stack_size`` (default: 128kB)
        and ``sys_check_interval`` (default: 10000). The former sets the
        default stack size to desired value, to avoid excessively large
        thread stacks -- typical operating system default is 8 MB, which
        adds up rather a lot for lots of server threads. The latter sets
        python's ``sys.setcheckinterval``; the default is to increase this
        to avoid unnecessarily frequent checks for python's GIL, global
        interpreter lock. In general we want each thread to complete as
        quickly as possible without making unnecessary checks."""
        cpconfig = cherrypy.config

        # Determine server local base.
        port = getattr(self.srvconfig, 'port', 8080)
        local_base = getattr(self.srvconfig, 'local_base', socket.gethostname())
        if local_base.find(':') == -1:
            local_base = '%s:%d' % (local_base, port)

        # Set default server configuration.
        cherrypy.log = Logger()

        cpconfig.update({'server.max_request_body_size': 0})
        cpconfig.update({'server.environment': 'production'})
        cpconfig.update({'server.socket_host': '0.0.0.0'})
        cpconfig.update({'server.socket_port': port})
        cpconfig.update({'server.socket_queue_size': 100})
        cpconfig.update({'server.thread_pool': 100})
        cpconfig.update({'tools.proxy.on': True})
        cpconfig.update({'tools.proxy.base': local_base})
        cpconfig.update({'tools.time.on': True})
        cpconfig.update({'engine.autoreload.on': False})
        cpconfig.update({'request.show_tracebacks': False})
        cpconfig.update({'request.methods_with_bodies': ("POST", "PUT", "DELETE")})
        _thread.stack_size(getattr(self.srvconfig, 'thread_stack_size', 128 * 1024))
        # read sys_check_interval from server config which sets in instructions
        # as we previously used sys.setcheckinterval
        interval = getattr(self.srvconfig, 'sys_check_interval', 10000)
        # set check interval in seconds for sys.setswitchinterval
        sys.setswitchinterval(interval/1000)
        self.silent = getattr(self.srvconfig, 'silent', False)

        # Apply any override options from app config file.
        for section in ('engine', 'hooks', 'log', 'request', 'response',
                        'server', 'tools', 'wsgi', 'checker'):
            if not hasattr(self.srvconfig, section):
                continue
            for opt, value in viewitems(getattr(self.srvconfig, section).dictionary_()):
                if isinstance(value, ConfigSection):
                    for xopt, xvalue in viewitems(value.dictionary_()):
                        cpconfig.update({"%s.%s.%s" % (section, opt, xopt): xvalue})
                elif isinstance(value, str) or isinstance(value, int):
                    cpconfig.update({"%s.%s" % (section, opt): value})
                else:
                    raise RuntimeError("%s.%s should be string or int, got %s"
                                       % (section, opt, type(value)))

        # Apply security customisation.
        if hasattr(self.srvconfig, 'authz_defaults'):
            defsec = self.srvconfig.authz_defaults
            cpconfig.update({'tools.cms_auth.on': True})
            cpconfig.update({'tools.cms_auth.role': defsec['role']})
            cpconfig.update({'tools.cms_auth.group': defsec['group']})
            cpconfig.update({'tools.cms_auth.site': defsec['site']})

        if hasattr(self.srvconfig, 'authz_policy'):
            cpconfig.update({'tools.cms_auth.policy': self.srvconfig.authz_policy})
        cherrypy.log("INFO: final CherryPy configuration: %s" % pformat(cpconfig))

    def install_application(self):
        """Install application and its components from the configuration."""
        index = self.srvconfig.index

        # First instantiate non-view extensions.
        if getattr(self.config, 'extensions', None):
            for ext in self.config.extensions:
                name = ext._internal_name
                if not self.silent:
                    cherrypy.log("INFO: instantiating extension %s" % name)
                module_name, class_name = ext.object.rsplit(".", 1)
                module = __import__(module_name, globals(), locals(), [class_name])
                obj = getattr(module, class_name)(self, ext)
                self.extensions[name] = obj

        # Then instantiate views and mount them to cherrypy. If the view is
        # designated as the index, create it as an application, profiled one
        # if server profiling was requested. Otherwise just mount it as a
        # normal server content object. Force tracebacks off for everything.
        for view in self.config.views:
            name = view._internal_name
            path = "/%s" % self.appname + ((name != index and "/%s" % name) or "")
            if not self.silent:
                cherrypy.log("INFO: loading %s into %s" % (name, path))
            module_name, class_name = view.object.rsplit(".", 1)
            module = __import__(module_name, globals(), locals(), [class_name])
            obj = getattr(module, class_name)(self, view, path)
            app = Application(obj, path, {"/": {"request.show_tracebacks": False}})
            if getattr(self.srvconfig, 'profile', False):
                profdir = "%s/profile" % self.statedir
                if not os.path.exists(profdir):
                    os.makedirs(profdir)
                app = ProfiledApp(app, profdir)
            cherrypy.tree.mount(app)
            self.views[name] = obj


class RESTDaemon(RESTMain):
    """Web server object.

    The `RESTDaemon` represents the web server daemon. It provides all
    services for starting, stopping and checking the status of the daemon,
    as well as running the main loop.

    The class implements all the methods required for proper unix
    daemonisation, including maintaing a PID file for the process group
    and correct progressively more aggressive signals sent to terminate
    the daemon. Starting multiple daemons in same directory is refused.

    The daemon takes the server configuration as a parametre. When the
    server is started, it creates a CherryPy server and configuration
    from the application config contents."""

    def __init__(self, config, statedir):
        """Initialise the daemon.

        :arg config: server configuration
        :arg str statedir: server state directory."""
        RESTMain.__init__(self, config, statedir)
        self.pidfile = "%s/%s.pid" % (self.statedir, self.appname)
        self.logfile = ["rotatelogs", "%s/%s-%%Y%%m%%d.log" % (self.statedir, self.appname), "86400"]

    def daemon_pid(self):
        """Check if there is a daemon running, and if so return its pid.

        Reads the pid file from the daemon work directory, if any. If a
        non-empty pid file exists, checks if a process by that PGID exists.
        If yes, reports the daemon as running, otherwise reports either a
        stale daemon no longer running or no deamon running at all.

        :returns: A tuple (running, pgid). The first value will be True if a
          running daemon was found, in which case pid will be its PGID. The
          first value will be false otherwise, and pgid will be either None
          if no pid file was found, or an integer if there was a stale file."""
        pid = None
        try:
            with open(self.pidfile) as fd:
                pid = int(fd.readline())
            os.killpg(pid, 0)
            return (True, pid)
        except:
            return (False, pid)

    def kill_daemon(self, silent=False):
        """Check if the daemon is running, and if so kill it.

        If there is no daemon running and no pid file, does nothing. If there
        is a pid file but no such process running, removes the stale pid file.
        Otherwise sends progressively more lethal signals at intervals to the
        daemon process until it quits.

        The signals are always sent to the entire process group, and signals
        will keep on getting sent as long as at least one process from the
        daemon process group is still alive. If for some reason the group
        cannot be killed otherwise, sends SIGKILL to the group in the end.

        The message about removing a stale pid file cannot be silenced. All
        other messages are squelched if `silent` is True.

        :arg bool silent: do not print any messages if True."""
        running, pid = self.daemon_pid()
        if not running:
            if pid != None:
                print("Removing stale pid file %s" % self.pidfile)
                os.remove(self.pidfile)
            else:
                if not silent:
                    print("%s not running, not killing" % self.appname)
        else:
            if not silent:
                sys.stdout.write("Killing %s pgid %d " % (self.appname, pid))
                sys.stdout.flush()

            dead = False
            for sig, grace in ((signal.SIGINT, .5), (signal.SIGINT, 1),
                               (signal.SIGINT, 3), (signal.SIGINT, 5),
                               (signal.SIGKILL, 0)):
                try:
                    if not silent:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    os.killpg(pid, sig)
                    time.sleep(grace)
                    os.killpg(pid, 0)
                except OSError as e:
                    if e.errno == errno.ESRCH:
                        dead = True
                        break
                    raise

            if not dead:
                try:
                    os.killpg(pid, signal.SIGKILL)
                except:
                    pass

            if not silent:
                sys.stdout.write("\n")

    def start_daemon(self):
        """Start the deamon."""
        # This REST server can be run in the frontend i.e. interactively (e.g. to use pdb)
        # by setting an env. var. via: export DONT_DAEMONIZE_REST=True
        # if the variable is missing or set to any other value, code starts normally as a daemon
        daemonize = os.getenv('DONT_DAEMONIZE_REST', 'False') != 'True'

        os.chdir(self.statedir)

        if daemonize:
            # Redirect all output to the logging daemon.
            devnull = open(os.devnull, "w")
            if isinstance(self.logfile, list):
                subproc = Popen(self.logfile, stdin=PIPE, stdout=devnull, stderr=devnull,
                                bufsize=0, close_fds=True, shell=False)
                logger = subproc.stdin
            elif isinstance(self.logfile, str):
                # if a unix pipe is set as the logfile, it must be opened to append to the end of the file
                # if file/pipe does not exist, create it
                logger = open(self.logfile, "a")
            else:
                raise TypeError("'logfile' must be a string or array")
            os.dup2(logger.fileno(), sys.stdout.fileno())
            os.dup2(logger.fileno(), sys.stderr.fileno())
            os.dup2(devnull.fileno(), sys.stdin.fileno())
            logger.close()
            devnull.close()

            # First fork. Discard the parent.
            pid = os.fork()
            if pid > 0:
                os._exit(0)

            # Establish as a daemon, set process group / session id.
            os.setsid()

            # Second fork. The child does the work, discard the second parent.
            pid = os.fork()
            if pid > 0:
                os._exit(0)

            # Save process group id to pid file, then run real worker.
            with open(self.pidfile, "w") as pidObj:
                pidObj.write("%d\n" % os.getpgid(0))

        # following code is executed both in daemon and not daemon mode
        error = False
        try:
            self.run()
        except Exception as e:
            error = True
            if PY2:
                trace = BytesIO()
            else:
                trace = StringIO()
            traceback.print_exc(file=trace)
            cherrypy.log("ERROR: terminating due to error: %s" % trace.getvalue())

        # Remove pid file once we are done.
        try:
            os.remove(self.pidfile)
        except:
            pass

        # Exit
        sys.exit((error and 1) or 0)

    def run(self):
        """Run the server daemon main loop."""
        # Fork.  The child always exits the loop and executes the code below
        # to run the server proper.  The parent monitors the child, and if
        # it exits abnormally, restarts it, otherwise exits completely with
        # the child's exit code.
        daemonize = os.getenv('DONT_DAEMONIZE_REST', 'False') != 'True'
        if daemonize:
            cherrypy.log("WATCHDOG: starting server daemon (pid %d)" % os.getpid())
            while True:
                serverpid = os.fork()
                if not serverpid: break
                signal.signal(signal.SIGINT, signal.SIG_IGN)
                signal.signal(signal.SIGTERM, signal.SIG_IGN)
                signal.signal(signal.SIGQUIT, signal.SIG_IGN)
                (xpid, exitrc) = os.waitpid(serverpid, 0)
                (exitcode, exitsigno, exitcore) = (exitrc >> 8, exitrc & 127, exitrc & 128)
                retval = (exitsigno and ("signal %d" % exitsigno)) or str(exitcode)
                retmsg = retval + ((exitcore and " (core dumped)") or "")
                restart = (exitsigno > 0 and exitsigno not in (2, 3, 15))
                cherrypy.log("WATCHDOG: server exited with exit code %s%s"
                             % (retmsg, (restart and "... restarting") or ""))

                if not restart:
                    sys.exit((exitsigno and 1) or exitcode)

                for pidfile in glob("%s/*/*pid" % self.statedir):
                    if os.path.exists(pidfile):
                        with open(pidfile) as fd:
                            pid = int(fd.readline())
                        os.remove(pidfile)
                        cherrypy.log("WATCHDOG: killing slave server %d" % pid)
                        try:
                            os.kill(pid, 9)
                        except:
                            pass

        # Run. Override signal handlers after CherryPy has itself started and
        # installed its own handlers. To achieve this we need to start the
        # server in non-blocking mode, fiddle with, than ask server to block.
        self.validate_config()
        self.setup_server()
        self.install_application()
        cherrypy.log("INFO: starting server in %s" % self.statedir)
        cherrypy.config.update({'log.screen': bool(getattr(self.srvconfig, "log_screen", True))})
        cherrypy.engine.start()
        signal.signal(signal.SIGHUP, sig_reload)
        signal.signal(signal.SIGUSR1, sig_graceful)
        signal.signal(signal.SIGTERM, sig_terminate)
        signal.signal(signal.SIGQUIT, sig_terminate)
        signal.signal(signal.SIGINT, sig_terminate)
        cherrypy.engine.block()


def main():
    # Re-exec if we don't have unbuffered i/o. This is essential to get server
    # to output its logs synchronous to its operation, such that log output does
    # not remain buffered in the python server. This is particularly important
    # when infrequently accessed server redirects output to 'rotatelogs'.
    if 'PYTHONUNBUFFERED' not in os.environ:
        os.environ['PYTHONUNBUFFERED'] = "1"
        if PY2:
            os.execvp("python", ["python"] + sys.argv)
        else:
            os.execvp("python3", ["python3"] + sys.argv)

    opt = ArgumentParser(usage=__doc__)
    opt.add_argument("-q", "--quiet", action="store_true", dest="quiet", default=False,
                     help="be quiet, don't print unnecessary output")
    opt.add_argument("-v", "--verify", action="store_true", dest="verify", default=False,
                     help="verify daemon is running, restart if not")
    opt.add_argument("-s", "--status", action="store_true", dest="status", default=False,
                     help="check if the server monitor daemon is running")
    opt.add_argument("-k", "--kill", action="store_true", dest="kill", default=False,
                     help="kill any existing already running daemon")
    opt.add_argument("-r", "--restart", action="store_true", dest="restart", default=False,
                     help="restart, kill any existing running daemon first")
    opt.add_argument("-d", "--dir", dest="statedir", metavar="DIR", default=os.getcwd(),
                     help="server state directory (default: current working directory)")
    opt.add_argument("-l", "--log", dest="logfile", metavar="DEST", default=None,
                     help="log to DEST, via pipe if DEST begins with '|', otherwise a file")
    opts, args = opt.parse_known_args()

    if len(args) != 1:
        print("%s: exactly one configuration file required" % sys.argv[0], file=sys.stderr)
        sys.exit(1)

    if not os.path.isfile(args[0]) or not os.access(args[0], os.R_OK):
        print("%s: %s: invalid configuration file" % (sys.argv[0], args[0]), file=sys.stderr)
        sys.exit(1)

    if not opts.statedir or \
            not os.path.isdir(opts.statedir) or \
            not os.access(opts.statedir, os.W_OK):
        print("%s: %s: invalid state directory" % (sys.argv[0], opts.statedir), file=sys.stderr)
        sys.exit(1)

    # Create server object.
    cfg = loadConfigurationFile(args[0])
    app = cfg.main.application.lower()
    server = RESTDaemon(cfg, opts.statedir)

    # Now actually execute the task.
    if opts.status:
        # Show status of running daemon, including exit code matching the
        # daemon status: 0 = running, 1 = not running, 2 = not running but
        # there is a stale pid file. If silent don't print out anything
        # but still return the right exit code.
        running, pid = server.daemon_pid()
        if running:
            if not opts.quiet:
                print("%s is %sRUNNING%s, PID %d" \
                      % (app, COLOR_OK, COLOR_NORMAL, pid))
            sys.exit(0)
        elif pid != None:
            if not opts.quiet:
                print("%s is %sNOT RUNNING%s, stale PID %d" \
                      % (app, COLOR_WARN, COLOR_NORMAL, pid))
            sys.exit(2)
        else:
            if not opts.quiet:
                print("%s is %sNOT RUNNING%s" \
                      % (app, COLOR_WARN, COLOR_NORMAL))
            sys.exit(1)

    elif opts.kill:
        # Stop any previously running daemon. If quiet squelch messages,
        # except removal of stale pid file cannot be silenced.
        server.kill_daemon(silent=opts.quiet)

    else:
        # We are handling a server start, in one of many possible ways:
        # normal start, restart (= kill any previous daemon), or verify
        # (= if daemon is running leave it alone, otherwise start).

        # Convert 'verify' to 'restart' if the server isn't running.
        if opts.verify:
            opts.restart = True
            if server.daemon_pid()[0]:
                sys.exit(0)

        # If restarting, kill any previous daemon, otherwise complain if
        # there is a daemon already running here. Starting overlapping
        # daemons is not supported because pid file would be overwritten
        # and we'd lose track of the previous daemon.
        if opts.restart:
            server.kill_daemon(silent=opts.quiet)
        else:
            running, pid = server.daemon_pid()
            if running:
                print("Refusing to start over an already running daemon, pid %d" % pid, file=sys.stderr)
                sys.exit(1)

        # If we are (re)starting and were given a log file option, convert
        # the logfile option to a list if it looks like a pipe request, i.e.
        # starts with "|", such as "|rotatelogs foo/bar-%Y%m%d.log".
        if opts.logfile:
            if opts.logfile.startswith("|"):
                server.logfile = re.split(r"\s+", opts.logfile[1:])
            else:
                server.logfile = opts.logfile

        # Actually start the daemon now.
        server.start_daemon()



#!/usr/bin/env python
"""
File       : MSPileupTaskManager.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileupTaskManager handles MSPileupTasks

In particular, it perform the following tasks each polling cycle:
    - fetches pileup sizes for all pileup documents in database back-end
    - update RSE quotas
    - perform monitoring task
    - perform task for active pileups using up-to-date RSE quotas
    - perform task for inactive pileups
"""

# system modules
from threading import current_thread

# WMCore modules
from Utils.Timers import CodeTimer
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.MicroService.DataStructs.DefaultStructs import PILEUP_REPORT
from WMCore.MicroService.MSPileup.MSPileupData import MSPileupData
from WMCore.MicroService.MSPileup.MSPileupTasks import MSPileupTasks
from WMCore.MicroService.MSPileup.MSPileupMonitoring import MSPileupMonitoring


class MSPileupTaskManager(MSCore):
    """
    MSPileupTaskManager handles MSPileup tasks
    """

    def __init__(self, msConfig, **kwargs):
        super().__init__(msConfig, **kwargs)
        self.marginSpace = msConfig.get('marginSpace', 1024**4)
        self.rucioAccount = msConfig.get('rucioAccount', 'wmcore_pileup')
        self.rucioUrl = msConfig['rucioUrl']  # aligned with MSCore init
        self.rucioAuthUrl = msConfig['rucioAuthUrl']  # aligned with MSCore init
        self.cleanupDaysThreshold = msConfig.get('cleanupDaysThreshold', 15)
        dryRun = msConfig.get('dryRun', False)
        rucioScope = msConfig.get('rucioScope', 'cms')
        customRucioScope = msConfig.get('customRucioScope', 'group.wmcore')
        self.rucioClient = self.rucio  # set in MSCore init
        self.dataManager = MSPileupData(msConfig)
        self.monitManager = MSPileupMonitoring(msConfig)
        self.mgr = MSPileupTasks(self.dataManager, self.monitManager, self.logger,
                                 self.rucioAccount, self.rucioClient, rucioScope, customRucioScope, dryRun)

    def status(self):
        """
        Provide MSPileupTaskManager status API.

        :return: status dictionary
        """
        summary = dict(PILEUP_REPORT)
        summary.update({'thread_id': current_thread().name})
        report = self.mgr.getReport()
        summary.update({'tasks': report.getDocuments()})
        return summary

    def executeCycle(self):
        """
        execute MSPileupTasks polling cycle
        """
        with CodeTimer("### pileupSizeTask", logger=self.logger):
            self.mgr.pileupSizeTask()
        with CodeTimer("### monitoringTask", logger=self.logger):
            self.mgr.monitoringTask()
        with CodeTimer("### activeTask", logger=self.logger):
            self.mgr.activeTask(marginSpace=self.marginSpace)
        with CodeTimer("### inactiveTask", logger=self.logger):
            self.mgr.inactiveTask()
        with CodeTimer("### partialPileupTask", logger=self.logger):
            self.mgr.partialPileupTask()
        with CodeTimer("### cleanupTask", logger=self.logger):
            self.mgr.cleanupTask(self.cleanupDaysThreshold)
        with CodeTimer("### cmsMonitTask", logger=self.logger):
            self.mgr.cmsMonitTask()



#!/usr/bin/env python3
# coding=utf-8
"""
Rucio Service class developed on top of the native Rucio Client
APIs providing custom output and handling as necessary for the
CMS Workload Management system (and migration from PhEDEx).
"""
from __future__ import division, print_function, absolute_import

from builtins import str, object
from future.utils import viewitems, viewvalues

import json
import logging
from copy import deepcopy
from rucio.client import Client
from rucio.common.exception import (AccountNotFound, DataIdentifierNotFound, AccessDenied, DuplicateRule,
                                    DataIdentifierAlreadyExists, DuplicateContent, InvalidRSEExpression,
                                    UnsupportedOperation, FileAlreadyExists, RuleNotFound, RSENotFound)
from Utils.MemoryCache import MemoryCache
from Utils.IteratorTools import grouper
from WMCore.Services.Rucio.RucioUtils import (validateMetaData, weightedChoice,
                                              isTapeRSE, dropTapeRSEs)
from WMCore.WMException import WMException


class WMRucioException(WMException):
    """
    _WMRucioException_
    Generic WMCore exception for Rucio
    """
    pass


class WMRucioDIDNotFoundException(WMException):
    """
    _WMRucioDIDNotFoundException_
    Generic WMCore exception for Rucio
    """
    pass


class Rucio(object):
    """
    Service class providing additional Rucio functionality on top of the
    Rucio client APIs.

    Just a reminder, we have different naming convention between Rucio
    and CMS, where:
     * CMS dataset corresponds to a Rucio container
     * CMS block corresponds to a Rucio dataset
     * CMS file corresponds to a Rucio file
    We will try to use Container -> CMS dataset, block -> CMS block.
    """

    def __init__(self, acct, hostUrl=None, authUrl=None, configDict=None, client=None):
        """
        Constructs a Rucio object with the Client object embedded.
        In order to instantiate a Rucio Client object, it assumes the host has
        a proper rucio configuration file, where the default host and authentication
        host URL come from, as well as the X509 certificate information.
        :param acct: rucio account to be used
        :param hostUrl: defaults to the rucio config one
        :param authUrl: defaults to the rucio config one
        :param configDict: dictionary with extra parameters
        :param client: optional Rucio client to use (useful for mock-up)
        """
        configDict = configDict or {}
        # default RSE data caching to 12h
        rseCacheExpiration = configDict.pop('cacheExpiration', 12 * 60 * 60)
        self.logger = configDict.pop("logger", logging.getLogger())

        self.rucioParams = deepcopy(configDict)
        self.rucioParams.setdefault('account', acct)
        self.rucioParams.setdefault('rucio_host', hostUrl)
        self.rucioParams.setdefault('auth_host', authUrl)
        self.rucioParams.setdefault('ca_cert', None)
        self.rucioParams.setdefault('auth_type', None)
        self.rucioParams.setdefault('creds', None)
        self.rucioParams.setdefault('timeout', 600)
        self.rucioParams.setdefault('user_agent', 'wmcore-client')

        self.logger.info("WMCore Rucio initialization parameters: %s", self.rucioParams)
        if client:
            self.cli = client
        else:
            self.cli = Client(rucio_host=hostUrl, auth_host=authUrl, account=acct,
                              ca_cert=self.rucioParams['ca_cert'], auth_type=self.rucioParams['auth_type'],
                              creds=self.rucioParams['creds'], timeout=self.rucioParams['timeout'],
                              user_agent=self.rucioParams['user_agent'])
            clientParams = {}
            for k in ("host", "auth_host", "auth_type", "account", "user_agent",
                      "ca_cert", "creds", "timeout", "request_retries"):
                clientParams[k] = getattr(self.cli, k)
            self.logger.info("Rucio client initialization parameters: %s", clientParams)

        # keep a map of rse expression to RSE names mapped for some time
        self.cachedRSEs = MemoryCache(rseCacheExpiration, {})

    def pingServer(self):
        """
        _pingServer_

        Ping the rucio server to see whether it's alive
        :return: a dict with the server version
        """
        return self.cli.ping()

    def whoAmI(self):
        """
        _whoAmI_

        Get information about account whose token is used
        :return: a dict with the account information. None in case of failure
        """
        return self.cli.whoami()

    def getAccount(self, acct):
        """
        _getAccount_

        Gets information about a specific account
        :return: a dict with the account information. None in case of failure
        """
        res = None
        try:
            res = self.cli.get_account(acct)
        except (AccountNotFound, AccessDenied) as ex:
            self.logger.error("Failed to get account information from Rucio. Error: %s", str(ex))
        return res

    def getAccountLimits(self, acct):
        """
        Provided an account name, fetch the storage quota for all RSEs
        :param acct: a string with the rucio account name
        :return: a dictionary of RSE name and quota in bytes.
        """
        res = {}
        try:
            res = self.cli.get_local_account_limits(acct)
        except AccountNotFound as ex:
            self.logger.error("Account: %s not found in the Rucio Server. Error: %s", acct, str(ex))
        return res

    def getAccountUsage(self, acct, rse=None):
        """
        _getAccountUsage_

        Provided an account name, gets the storage usage for it against
        a given RSE (or all RSEs)
        :param acct: a string with the rucio account name
        :param rse: an optional string with the RSE name
        :return: a list of dictionaries with the account usage information.
          None in case of failure
        """
        res = []
        try:
            res = list(self.cli.get_local_account_usage(acct, rse=rse))
        except (AccountNotFound, AccessDenied) as ex:
            self.logger.error("Failed to get account usage information from Rucio. Error: %s", str(ex))
        return res

    def getBlocksInContainer(self, container, scope='cms'):
        """
        _getBlocksInContainer_

        Provided a Rucio container - CMS dataset - retrieve all blocks in it.
        :param container: a CMS dataset string
        :param scope: string containing the Rucio scope (defaults to 'cms')
        :return: a list of block names
        """
        blockNames = []
        if not self.isContainer(container, scope=scope):
            # input container wasn't really a container
            self.logger.warning("Provided DID name is not a CONTAINER type: %s", container)
            return blockNames

        response = self.cli.list_content(scope=scope, name=container)
        for item in response:
            if item['type'].upper() == 'DATASET':
                blockNames.append(item['name'])

        return blockNames

    def getReplicaInfoForBlocks(self, **kwargs):
        """
        _getReplicaInfoForBlocks_

        Get block replica information.
        It used to mimic the same PhEDEx wrapper API, listing all the
        current locations where the provided input data is.

        :param kwargs:  Supported keyword arguments:
           * block:    List of either dataset or block names. Not both!,
           * deep:     Whether or not to lookup for replica info at file level
        :return: a list of dictionaries with replica information
        """
        kwargs.setdefault("scope", "cms")
        kwargs.setdefault("deep", False)

        blockNames = []
        result = []

        if isinstance(kwargs.get('block', None), (list, set)):
            blockNames = kwargs['block']
        elif 'block' in kwargs:
            blockNames = [kwargs['block']]

        if isinstance(kwargs.get('dataset', None), (list, set)):
            for datasetName in kwargs['dataset']:
                blockNames.extend(self.getBlocksInContainer(datasetName, scope=kwargs['scope']))
        elif 'dataset' in kwargs:
            blockNames.extend(self.getBlocksInContainer(kwargs['dataset'], scope=kwargs['scope']))

        inputDids = []
        for block in blockNames:
            inputDids.append({"scope": kwargs["scope"], "type": "DATASET", "name": block})

        resultDict = {}
        if kwargs['deep']:
            for did in inputDids:
                for item in self.cli.list_dataset_replicas(kwargs["scope"], did["name"], deep=kwargs['deep']):
                    resultDict.setdefault(item['name'], [])
                    if item['state'].upper() == 'AVAILABLE':
                        resultDict[item['name']].append(item['rse'])
        else:
            for item in self.cli.list_dataset_replicas_bulk(inputDids):
                resultDict.setdefault(item['name'], [])
                if item['state'].upper() == 'AVAILABLE':
                    resultDict[item['name']].append(item['rse'])

        # Finally, convert it to a list of dictionaries, like:
        # [{"name": "block_A", "replica": ["nodeA", "nodeB"]},
        #  {"name": "block_B", etc etc}]
        for blockName, rses in viewitems(resultDict):
            result.append({"name": blockName, "replica": list(set(rses))})

        return result

    def getPFN(self, site, lfns, protocol=None, protocolDomain='ALL', operation=None):
        """
        returns a list of PFN(s) for a list of LFN(s) and one site
        Note: same function implemented for PhEDEx accepted a list of sites, but semantic was obscure and actual need
              unknown. So take this chance to make things simple.
        See here for documentation of the upstream Rucio API: https://rucio.readthedocs.io/en/latest/api/rse.html
        :param site: a Rucio RSE, i.e. a site name in standard CMS format like 'T1_UK_RAL_Disk' or  'T2_CH_CERN'
        :param lfns: one LFN or a list of LFN's, does not need to correspond to actual files and could be a top level directory
                      like ['/store/user/rucio/jdoe','/store/data',...] or the simple '/store/data'
        :param protocol: If the RSE supports multiple access protocols, a preferred protocol can be selected via this,
                         otherwise the default one for the site will be selected. Example: 'gsiftp' or 'davs'
                         this is what is called "scheme" in the RUCIO API (we think protocol is more clear)
        :param protocolDomain: The scope of the protocol. Supported are "LAN", "WAN", and "ALL" (as default)
                                this is what is called protocol_domain in RUCIO API, changed for CMS camelCase convention
        :param operation: The name of the requested operation (read, write, or delete). If None, all operations are queried
        :return: a dictionary having the LFN's as keys and the corresponding PFN's as values.

        Will raise a Rucio exception if input is wrong.
        """

        # allow for duck typing, if a single lfn was passed, make it a list
        # avoid handling a string like a list of chars !
        if not isinstance(lfns, (set, list)):
            lfns = [lfns]

        # add a scope to turn LFNs into Rucio DID syntax
        dids = ['cms:' + lfn for lfn in lfns]

        # Rucio's lfns2pfns returns a dictionary with did as key and pfn as value:
        # {u'cms:/store/user/rucio': u'gsiftp://red-gridftp.unl.edu:2811/mnt/hadoop/user/uscms01/pnfs/unl.edu/data4/cms/store/user/rucio'}
        didDict = self.cli.lfns2pfns(site, dids, scheme=protocol, protocol_domain=protocolDomain, operation=operation)

        # convert to a more useful format for us with LFN's as keys
        pfnDict = {}
        for oldKey in didDict:
            newKey = oldKey.lstrip('cms:')
            pfnDict[newKey] = didDict[oldKey]

        return pfnDict

    def createContainer(self, name, scope='cms', **kwargs):
        """
        _createContainer_

        Create a CMS dataset (Rucio container) in a given scope.
        :param name: string with the container name
        :param scope: optional string with the scope name
        :param kwargs:  supported keyword arguments (from the Rucio CLI API documentation):
           * statuses:  Dictionary with statuses, e.g.g {"monotonic":True}.
           * meta:      Meta-data associated with the data identifier is represented using key/value
                        pairs in a dictionary.
           * rules:     Replication rules associated with the data identifier. A list of dictionaries,
                        e.g., [{"copies": 2, "rse_expression": "TIERS1"}, ].
           * lifetime:  DID's lifetime (in seconds).
        :return: a boolean to represent whether it succeeded or not
        """
        response = False
        if not validateMetaData(name, kwargs.get("meta", {}), logger=self.logger):
            return response
        try:
            # add_container(scope, name, statuses=None, meta=None, rules=None, lifetime=None)
            response = self.cli.add_container(scope, name, **kwargs)
        except DataIdentifierAlreadyExists:
            self.logger.debug("Container name already exists in Rucio: %s", name)
            response = True
        except Exception as ex:
            self.logger.error("Exception creating container: %s. Error: %s", name, str(ex))
        return response

    def createBlock(self, name, scope='cms', attach=True, **kwargs):
        """
        _createBlock_

        Create a CMS block (Rucio dataset) in a given scope. It also associates
        the block to its container
        :param name: string with the block name
        :param scope: optional string with the scope name
        :param attach: boolean whether to attack this block to its container or not
        :param kwargs:  supported keyword arguments (from the Rucio CLI API documentation):
           * statuses:  Dictionary with statuses, e.g. {"monotonic":True}.
           * lifetime:  DID's lifetime (in seconds).
           * files:     The content.
           * rse:       The RSE name when registering replicas.
           * meta:      Meta-data associated with the data identifier. Represented
                        using key/value pairs in a dictionary.
           * rules:     replication rules associated with the data identifier. A list of
                        dictionaries, e.g., [{"copies": 2, "rse_expression": "TIERS1"}, ].
        :return: a boolean with the outcome of the operation

        NOTE: we will very likely define a rule saying: keep this block at the location it's
              being produced
        """
        response = False
        if not validateMetaData(name, kwargs.get("meta", {}), logger=self.logger):
            return response
        try:
            # add_dataset(scope, name, statuses=None, meta=None, rules=None, lifetime=None, files=None, rse=None)
            response = self.cli.add_dataset(scope, name, **kwargs)
        except DataIdentifierAlreadyExists:
            self.logger.debug("Block name already exists in Rucio: %s", name)
            response = True
        except Exception as ex:
            self.logger.error("Exception creating block: %s. Error: %s", name, str(ex))

        # then attach this block recently created to its container
        if response and attach:
            container = name.split('#')[0]
            response = self.attachDIDs(kwargs.get('rse'), container, name, scope)
        return response

    def attachDIDs(self, rse, superDID, dids, scope='cms', chunkSize=1000):
        """
        _attachDIDs_

        Create a list of files - in bulk - to a RSE. Then attach them to the block name.
        :param rse: string with the RSE name
        :param superDID: upper structure level (can be a container or block. If attaching blocks,
             then it's a container name; if attaching files, then it's a block name)
        :param dids: either a string or a list of data identifiers (can be block or files)
        :param scope: string with the scope name
        :param chunkSize: maximum number of dids to be attached in any single Rucio server call
        :return: a boolean to represent whether it succeeded or not
        """
        if not isinstance(dids, list):
            dids = [dids]
        # NOTE: the attaching dids do not create new container within a scope
        # and it is safe to use cms scope for it
        alldids = [{'scope': 'cms', 'name': did} for did in sorted(dids)]

        # report if we use chunk size in rucio attach_dids API call
        if len(alldids) > chunkSize:
            self.logger.info("Attaching a total of %d DIDs in chunk size of: %d", len(alldids), chunkSize)

        for dids in grouper(alldids, chunkSize):
            response = False
            try:
                response = self.cli.attach_dids(scope, superDID, dids=dids, rse=rse)
            except DuplicateContent:
                self.logger.warning("Dids: %s already attached to: %s", dids, superDID)
                response = True
            except FileAlreadyExists:
                self.logger.warning("Failed to attach files already existent on block: %s", superDID)
                response = True
            except DataIdentifierNotFound:
                self.logger.error("Failed to attach dids: %s. Parent DID %s does not exist.", dids, superDID)
            except Exception as ex:
                self.logger.error("Exception attaching %s dids to: %s. Error: %s. First 10 dids: %s",
                                  len(dids), superDID, str(ex), dids[:10])
            if not response:
                # if we had failure with specific chunk of dids we'll return immediately
                return response

        return response

    def createReplicas(self, rse, files, block, scope='cms', ignoreAvailability=True):
        """
        _createReplicas_

        Create a list of files - in bulk - to a RSE. Then attach them to the block name.
        :param rse: string with the RSE name
        :param files: list of dictionaries with the file names and some of its meta data.
            E.g.: {'name': lfn, 'bytes': size, 'scope': scope, 'adler32': checksum, 'state': 'A'}
            State 'A' means that the replica is available
        :param block: string with the block name
        :param scope: string with the scope name
        :param ignore_availability: boolean to ignore the RSE blacklisting
        :return: a boolean to represent whether it succeeded or not
        """
        if isinstance(files, dict):
            files = [files]
        for item in files:
            item['scope'] = scope

        response = False
        try:
            # add_replicas(rse, files, ignore_availability=True)
            response = self.cli.add_replicas(rse, files, ignoreAvailability)
        except DataIdentifierAlreadyExists as exc:
            if len(files) == 1:
                self.logger.debug("File replica already exists in Rucio: %s", files[0]['name'])
                response = True
            else:
                # FIXME: I think we would have to iterate over every single file and add then one by one
                errorMsg = "Failed to insert replicas for: {} and block: {}".format(files, block)
                errorMsg += " Some/all DIDs already exist in Rucio. Error: {}".format(str(exc))
                self.logger.error(errorMsg)
        except Exception as exc:
            self.logger.error("Failed to add replicas for: %s and block: %s. Error: %s", files, block, str(exc))

        if response:
            files = [item['name'] for item in files]
            response = self.attachDIDs(rse, block, files, scope)

        return response

    def closeBlockContainer(self, name, scope='cms'):
        """
        _closeBlockContainer_

        Method to close a block or container, such that it doesn't get any more
        blocks and/or files inserted.
        :param name: data identifier (either a block or a container name)
        :param scope: string with the scope name
        :return: a boolean to represent whether it succeeded or not
        """
        response = False
        try:
            response = self.cli.close(scope, name)
        except UnsupportedOperation:
            self.logger.warning("Container/block has been closed already: %s", name)
            response = True
        except DataIdentifierNotFound:
            self.logger.error("Failed to close DID: %s; it does not exist.", name)
        except Exception as ex:
            self.logger.error("Exception closing container/block: %s. Error: %s", name, str(ex))
        return response

    def createReplicationRule(self, names, rseExpression, scope='cms', copies=1, **kwargs):
        """
        _createReplicationRule_

        Creates a replication rule against a list of data identifiers.
        :param names: either a string with a did or a list of dids of the same type
          (either a block or a container name)
        :param rseExpression: boolean string expression to give the list of RSEs.
           Full documentation on: https://rucio.readthedocs.io/en/latest/rse_expressions.html
           E.g.: "tier=2&US=true"
        :param scope: string with the scope name
        :param kwargs:  supported keyword arguments (from the Rucio CLI API documentation):
           * weight:    If the weighting option of the replication rule is used,
             the choice of RSEs takes their weight into account.
           * lifetime:  The lifetime of the replication rules (in seconds).
           * grouping:  Decides how the replication is going to happen; where:
                ALL: All files will be replicated to the same RSE.
                DATASET: All files in the same block will be replicated to the same RSE.
                NONE: Files will be completely spread over all allowed RSEs without any
                grouping considerations at all.
                E.g.: ALL grouping for 3 blocks against 3 RSEs. All 3 blocks go to one of the RSEs
                      DATASET grouping for 3 blocks against 3 RSEs, each block gets fully replicated
                      to one of those random RSEs
           * account:   The account owning the rule.
           * locked:    If the rule is locked, it cannot be deleted.
           * source_replica_expression: RSE Expression for RSEs to be considered for source replicas.
           * activity:  Transfer Activity to be passed to FTS.
           * notify:    Notification setting for the rule (Y, N, C).
           * purge_replicas: When the rule gets deleted purge the associated replicas immediately.
           * ignore_availability: Option to ignore the availability of RSEs.
           * ask_approval: Ask for approval of this replication rule.
           * asynchronous: Create rule asynchronously by judge-injector.
           * priority:  Priority of the transfers.
           * comment:   Comment about the rule.
           * meta:      Metadata, as dictionary.
        :return: it returns either an empty list or a list with a string id for the rule created

        NOTE: if there is an AccessDenied rucio exception, it raises a WMRucioException
        """
        kwargs.setdefault('grouping', 'ALL')
        kwargs.setdefault('account', self.rucioParams.get('account'))
        kwargs.setdefault('lifetime', None)
        kwargs.setdefault('locked', False)
        kwargs.setdefault('notify', 'N')
        kwargs.setdefault('purge_replicas', False)
        kwargs.setdefault('ignore_availability', False)
        kwargs.setdefault('ask_approval', False)
        kwargs.setdefault('asynchronous', True)
        kwargs.setdefault('priority', 3)
        if isinstance(kwargs.get('meta'), dict):
            kwargs['meta'] = json.dumps(kwargs['meta'])

        if not isinstance(names, (list, set)):
            names = [names]
        dids = [{'scope': scope, 'name': did} for did in names]

        response = []
        try:
            response = self.cli.add_replication_rule(dids, copies, rseExpression, **kwargs)
        except AccessDenied as ex:
            msg = "AccessDenied creating DID replication rule. Error: %s" % str(ex)
            raise WMRucioException(msg)
        except DuplicateRule as ex:
            # NOTE:
            #    The unique constraint is per did and it is checked against the tuple:
            #    rucioAccount + did = (name, scope) + rseExpression
            # NOTE:
            #    This exception will be thrown by Rucio even if a single Did has
            #    a duplicate rule. In this case all the rest of the Dids will be
            #    ignored, which in general should be addressed by Rucio. But since
            #    it is not, we should break the list of Dids and proceed one by one
            self.logger.warning("Resolving duplicate rules and separating every DID in a new rule...")

            ruleIds = []
            # now try creating a new rule for every single DID in a separated call
            for did in dids:
                try:
                    response = self.cli.add_replication_rule([did], copies, rseExpression, **kwargs)
                    ruleIds.extend(response)
                except DuplicateRule:
                    self.logger.warning("Found duplicate rule for account: %s\n, rseExp: %s\ndids: %s",
                                        kwargs['account'], rseExpression, dids)
                    # Well, then let us find which rule_id is already in the system
                    for rule in self.listDataRules(did['name'], scope=did['scope'],
                                                   account=kwargs['account'], rse_expression=rseExpression):
                        ruleIds.append(rule['id'])
            return list(set(ruleIds))
        except Exception as ex:
            self.logger.error("Exception creating rule replica for data: %s. Error: %s", names, str(ex))
        return response

    def listRuleHistory(self, dids):
        """
        _listRuleHistory_

        A function to return a list of historical records of replication rules
        per did.
        :param dids: a list of dids of the form {'name': '...', 'scope: '...'}

        The returned structure looks something like:
        [{'did': {'name': 'DidName',
                  'scope': 'cms'},
          'did_hist': [{u'account': u'wma_test',
                        u'created_at': datetime.datetime(2020, 6, 30, 1, 34, 51),
                        u'locks_ok_cnt': 9,
                        u'locks_replicating_cnt': 0,
                        u'locks_stuck_cnt': 0,
                        u'rse_expression': u'(tier=2|tier=1)&cms_type=real&rse_type=DISK',
                        u'rule_id': u'1f0ab297e4b54e1abf7c086ac012b9e9',
                        u'state': u'OK',
                        u'updated_at': datetime.datetime(2020, 6, 30, 1, 34, 51)}]},
         ...
         {'did': {},
          'did_hist': []}]
        """
        fullHistory = []
        for did in dids:
            didHistory = {}
            didHistory['did'] = did
            didHistory['did_hist'] = []
            # check the full history of the current did
            for hist in self.cli.list_replication_rule_full_history(did['scope'], did['name']):
                didHistory['did_hist'].append(hist)
            fullHistory.append(didHistory)
        return fullHistory

    def listContent(self, name, scope='cms'):
        """
        _listContent_

        List the content of the data identifier, returning some very basic information.
        :param name: data identifier (either a block or a container name)
        :param scope: string with the scope name
        :return: a list with dictionary items
        """
        res = []
        try:
            res = self.cli.list_content(scope, name)
        except Exception as ex:
            self.logger.error("Exception listing content of: %s. Error: %s", name, str(ex))
        return list(res)

    def listDataRules(self, name, **kwargs):
        """
        _listDataRules_

        List all rules associated to the data identifier provided.
        :param name: data identifier (either a block or a container name)
        :param kwargs: key/value filters supported by list_replication_rules Rucio API, such as:
          * scope: string with the scope name (optional)
          * account: string with the rucio account name (optional)
          * state: string with the state name (optional)
          * grouping: string with the grouping name (optional)
          * did_type: string with the DID type (optional)
          * created_before: an RFC-1123 compliant date string (optional)
          * created_after: an RFC-1123 compliant date string (optional)
          * updated_before: an RFC-1123 compliant date string (optional)
          * updated_after: an RFC-1123 compliant date string (optional)
          * and any of the other supported query arguments from the ReplicationRule class, see:
          https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/models.py#L884
        :return: a list with dictionary items
        """
        kwargs["name"] = name
        kwargs.setdefault("scope", "cms")
        try:
            res = self.cli.list_replication_rules(kwargs)
        except Exception as exc:
            msg = "Exception listing rules for data: {} and kwargs: {}. Error: {}".format(name,
                                                                                          kwargs,
                                                                                          str(exc))
            raise WMRucioException(msg)
        return list(res)

    def listDataRulesHistory(self, name, scope='cms'):
        """
        _listDataRulesHistory_

        List the whole rule history of a given DID.
        :param name: data identifier (either a block or a container name)
        :param scope: string with the scope name
        :return: a list with dictionary items
        """
        res = []
        try:
            res = self.cli.list_replication_rule_full_history(scope, name)
        except Exception as ex:
            self.logger.error("Exception listing rules history for data: %s. Error: %s", name, str(ex))
        return list(res)

    def listParentDIDs(self, name, scope='cms'):
        """
        _listParentDID__

        List the parent block/container of the specified DID.
        :param name: data identifier (either a block or a file name)
        :param scope: string with the scope name
        :return: a list with dictionary items
        """
        res = []
        try:
            res = self.cli.list_parent_dids(scope, name)
        except Exception as ex:
            self.logger.error("Exception listing parent DIDs for data: %s. Error: %s", name, str(ex))
        return list(res)

    def getRule(self, ruleId):
        """
        _getRule_

        Retrieve rule information for a given rule id
        :param ruleId: string with the rule id
        :return: a dictionary with the rule data (or empty if no rule can be found)
        """
        res = {}
        try:
            res = self.cli.get_replication_rule(ruleId)
        except RuleNotFound:
            self.logger.error("Cannot find any information for rule id: %s", ruleId)
        except Exception as ex:
            self.logger.error("Exception getting rule id: %s. Error: %s", ruleId, str(ex))
        return res

    def updateRule(self, ruleId, opts):
        """
        Update rule information for a given rule id
        :param ruleId: string with the rule id
        :param opts: dictionary, rule id options passed to Rucio
        :return: boolean status of update call
        """
        status = None
        try:
            status = self.cli.update_replication_rule(ruleId, opts)
        except RuleNotFound:
            self.logger.error("Cannot find any information for rule id: %s", ruleId)
        except Exception as ex:
            self.logger.error("Exception updating rule id: %s. Error: %s", ruleId, str(ex))
        return status

    def deleteRule(self, ruleId, purgeReplicas=False):
        """
        _deleteRule_

        Deletes a replication rule and all its associated locks
        :param ruleId: string with the rule id
        :param purgeReplicas: bool,ean to immediately delete the replicas
        :return: a boolean to represent whether it succeeded or not
        """
        res = True
        try:
            res = self.cli.delete_replication_rule(ruleId, purge_replicas=purgeReplicas)
        except RuleNotFound:
            self.logger.error("Could not find rule id: %s. Assuming it has already been deleted", ruleId)
        except Exception as ex:
            self.logger.error("Exception deleting rule id: %s. Error: %s", ruleId, str(ex))
            res = False
        return res

    def evaluateRSEExpression(self, rseExpr, useCache=True, returnTape=True):
        """
        Provided an RSE expression, resolve it and return a flat list of RSEs
        :param rseExpr: an RSE expression (which could be the RSE itself...)
        :param useCache: boolean defining whether cached data is meant to be used or not
        :param returnTape: boolean to also return Tape RSEs from the RSE expression result
        :return: a list of RSE names
        """
        if self.cachedRSEs.isCacheExpired():
            self.cachedRSEs.reset()
        if useCache and rseExpr in self.cachedRSEs:
            if returnTape:
                return self.cachedRSEs[rseExpr]
            return dropTapeRSEs(self.cachedRSEs[rseExpr])
        else:
            matchingRSEs = []
            try:
                for item in self.cli.list_rses(rseExpr):
                    matchingRSEs.append(item['rse'])
            except InvalidRSEExpression as exc:
                msg = "Provided RSE expression is considered invalid: {}. Error: {}".format(rseExpr, str(exc))
                raise WMRucioException(msg)
        if useCache:
            # add this key/value pair to the cache
            self.cachedRSEs.addItemToCache({rseExpr: matchingRSEs})
        if returnTape:
            return matchingRSEs
        return dropTapeRSEs(matchingRSEs)

    def pickRSE(self, rseExpression='rse_type=TAPE\cms_type=test', rseAttribute='ddm_quota', minNeeded=0):
        """
        _pickRSE_

        Use a weighted random selection algorithm to pick an RSE for a dataset based on an attribute
        The attribute should correlate to space available.
        :param rseExpression: Rucio RSE expression to pick RSEs (defaults to production Tape RSEs)
        :param rseAttribute: The RSE attribute to use as a weight. Must be a number
        :param minNeeded: If the RSE attribute is less than this number, the RSE will not be considered.

        Returns: A tuple of the chosen RSE and if the chosen RSE requires approval to write (rule property)
        """
        matchingRSEs = self.evaluateRSEExpression(rseExpression)
        rsesWithApproval = []
        rsesWeight = []

        for rse in matchingRSEs:
            attrs = self.cli.list_rse_attributes(rse)
            if rseAttribute:
                try:
                    quota = float(attrs.get(rseAttribute, 0))
                except (TypeError, KeyError):
                    quota = 0
            else:
                quota = 1
            requiresApproval = attrs.get('requires_approval', False)
            if quota > minNeeded:
                rsesWithApproval.append((rse, requiresApproval))
                rsesWeight.append(quota)

        return weightedChoice(rsesWithApproval, rsesWeight)

    def requiresApproval(self, rse):
        """
        _requiresApproval_

        Returns wether or not the specified RSE requires rule approval.
        :param res: string containing the Rucio RSE name
        :returns: True if the RSE requires approval to write (rule property)
        """
        try:
            attrs = self.cli.list_rse_attributes(rse)
        except RSENotFound as exc:
            msg = "Error retrieving attributes for RSE: {}. Error: {}".format(rse, str(exc))
            raise WMRucioException(msg)
        return attrs.get('requires_approval', False)

    def isContainer(self, didName, scope='cms'):
        """
        Checks whether the DID name corresponds to a container type or not.
        :param didName: string with the DID name
        :param scope: string containing the Rucio scope (defaults to 'cms')
        :return: True if the DID is a container, else False
        """
        try:
            response = self.cli.get_did(scope=scope, name=didName)
        except DataIdentifierNotFound as exc:
            msg = "Data identifier not found in Rucio: {}. Error: {}".format(didName, str(exc))
            raise WMRucioDIDNotFoundException(msg)
        return response['type'].upper() == 'CONTAINER'

    def getDID(self, didName, dynamic=True, scope='cms'):
        """
        Retrieves basic information for a single data identifier.
        :param didName: string with the DID name
        :param dynamic: boolean to dynamically calculate the DID size (default to True)
        :param scope: string containing the Rucio scope (defaults to 'cms')
        :return: a dictionary with basic DID information
        """
        try:
            response = self.cli.get_did(scope=scope, name=didName, dynamic=dynamic)
        except DataIdentifierNotFound as exc:
            response = dict()
            self.logger.error("Data identifier not found in Rucio: %s. Error: %s", didName, str(exc))
        return response

    def didExist(self, didName, scope='cms'):
        """
        Provided a given DID, check whether it's already in the Rucio server.
        Any kind of exception will return False (thus, data not yet in Rucio).
        :param didName: a string with the DID name (container, block, or file)
        :return: True if DID has been found, False otherwise
        """
        try:
            response = self.cli.get_did(scope=scope, name=didName)
        except Exception:
            response = dict()
        return response.get("name") == didName

    # FIXME we can likely delete this method (replaced by another implementation)
    def getDataLockedAndAvailable_old(self, **kwargs):
        """
        This method retrieves all the locations where a given DID is
        currently available and locked. It can be used for the data
        location logic in global and local workqueue.

        Logic is as follows:
        1. look for all replication rules matching the provided keyword args
        2. location of single RSE rules and in state OK are set as a location
        3.a. if the input DID name is a block, get all the locks AVAILABLE for that block
          and compare the rule ID against the multi RSE rules. Keep the RSE if it matches
        3.b. otherwise - if it's a container - method `_getContainerLockedAndAvailable`
          gets called to resolve all the blocks and locks
        4. union of the single RSEs and the matched multi RSEs is returned as the
        final location of the provided DID

        :param kwargs: key/value pairs to filter out the rules. Most common filters are likely:
            scope: string with the scope name
            name: string with the DID name
            account: string with the rucio account name
            state: string with the state name
            grouping: string with the grouping name
        :param returnTape: boolean to return Tape RSEs in the output, if any
        :return: a flat list with the RSE names locking and holding the input DID

        NOTE: some of the supported values can be looked up at:
        https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/constants.py#L184
        """
        msg = "This method `getDataLockedAndAvailable_old` is getting deprecated "
        msg += "and it will be removed in future releases."
        self.logger.warning(msg)
        returnTape = kwargs.pop("returnTape", False)
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getDataLockedAndAvailable API")
        if 'grouping' in kwargs:
            # long strings seem not to be working, like ALL / DATASET
            if kwargs['grouping'] == "ALL":
                kwargs['grouping'] = "A"
            elif kwargs['grouping'] == "DATASET":
                kwargs['grouping'] = "D"

        kwargs.setdefault("scope", "cms")

        multiRSERules = []
        finalRSEs = set()

        # First, find all the rules and where data is supposed to be locked
        rules = self.cli.list_replication_rules(kwargs)
        for rule in rules:
            # now resolve the RSE expressions
            rses = self.evaluateRSEExpression(rule['rse_expression'], returnTape=returnTape)
            if rule['copies'] == len(rses) and rule['state'] == "OK":
                # then we can guarantee that data is locked and available on these RSEs
                finalRSEs.update(set(rses))
            else:
                multiRSERules.append(rule['id'])
        self.logger.debug("Data location for %s from single RSE locks and available at: %s",
                          kwargs['name'], list(finalRSEs))
        if not multiRSERules:
            # then that is it, we can return the current RSEs holding and locking this data
            return list(finalRSEs)

        # At this point, we might already have some of the RSEs where the data is available and locked
        # Now check dataset locks and compare those rules against our list of multi RSE rules
        if self.isContainer(kwargs['name'], scope=kwargs.get('scope', 'cms')):
            # It's a container! Find what those RSEs are and add them to the finalRSEs set
            rseLocks = self._getContainerLockedAndAvailable(multiRSERules, returnTape=returnTape, **kwargs)
            self.logger.debug("Data location for %s from multiple RSE locks and available at: %s",
                              kwargs['name'], list(rseLocks))
        else:
            # It's a single block! Find what those RSEs are and add them to the finalRSEs set
            rseLocks = set()
            for blockLock in self.cli.get_dataset_locks(kwargs['scope'], kwargs['name']):
                if blockLock['state'] == 'OK' and blockLock['rule_id'] in multiRSERules:
                    rseLocks.add(blockLock['rse'])
            self.logger.debug("Data location for %s from multiple RSE locks and available at: %s",
                              kwargs['name'], list(rseLocks))

        finalRSEs = list(finalRSEs | rseLocks)
        return finalRSEs

    # FIXME we can likely delete this method
    def _getContainerLockedAndAvailable_old(self, multiRSERules, **kwargs):
        """
        This method is only supposed to be called internally (private method),
        because it won't consider the container level rules.

        This method retrieves all the locations where a given container DID is
        currently available and locked. It can be used for the data
        location logic in global and local workqueue.

        Logic is as follows:
        1. find all the blocks in the provided container name
        2. loop over all the block names and fetch their locks AVAILABLE
        2.a. compare the rule ID against the multi RSE rules. Keep the RSE if it matches
        3.a. if keyword argument grouping is ALL, returns an intersection of all blocks RSEs
        3.b. else - grouping DATASET - returns an union of all blocks RSEs

        :param multiRSERules: list of container level rules to be matched against
        :param kwargs: key/value pairs to filter out the rules. Most common filters are likely:
            scope: string with the scope name
            name: string with the DID name
            account: string with the rucio account name
            state: string with the state name
            grouping: string with the grouping name
        :param returnTape: boolean to return Tape RSEs in the output, if any
        :return: a set with the RSE names locking and holding this input DID

        NOTE: some of the supported values can be looked up at:
        https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/constants.py#L184
        """
        msg = "This method `_getContainerLockedAndAvailable_old` is getting deprecated "
        msg += "and it will be removed in future releases."
        self.logger.warning(msg)
        returnTape = kwargs.pop("returnTape", False)
        finalRSEs = set()
        blockNames = self.getBlocksInContainer(kwargs['name'])
        self.logger.debug("Container: %s contains %d blocks. Querying dataset_locks ...",
                          kwargs['name'], len(blockNames))

        rsesByBlocks = {}
        for block in blockNames:
            rsesByBlocks.setdefault(block, set())
            # FIXME: feature request made to the Rucio team to support bulk operations:
            # https://github.com/rucio/rucio/issues/3982
            for blockLock in self.cli.get_dataset_locks(kwargs['scope'], block):
                if not returnTape and isTapeRSE(blockLock['rse']):
                    continue
                if blockLock['state'] == 'OK' and blockLock['rule_id'] in multiRSERules:
                    rsesByBlocks[block].add(blockLock['rse'])

        # The question now is:
        #   1. do we want to have a full copy of the container in the same RSEs (grouping=A)
        #   2. or we want all locations holding at least one block of the container (grouping=D)
        if kwargs.get('grouping') == 'A':
            firstRun = True
            for _block, rses in viewitems(rsesByBlocks):
                if firstRun:
                    finalRSEs = rses
                    firstRun = False
                else:
                    finalRSEs = finalRSEs & rses
        else:
            for _block, rses in viewitems(rsesByBlocks):
                finalRSEs = finalRSEs | rses
        return finalRSEs

    def getPileupLockedAndAvailable(self, container, account, scope="cms"):
        """
        Method to resolve where the pileup container (and all its blocks)
        is locked and available.

        Pileup location resolution involves the following logic:
        1. find replication rules at the container level
          * if num of copies is equal to num of rses, and state is Ok, use
          those RSEs as container location (thus, every single block)
          * elif there are more rses than copies, keep that rule id for the next step
        2. discover all the blocks in the container
        3. if there are no multi RSEs rules, just build the block location map and return
        3. otherwise, for every block, list their current locks and if they are in state=OK
           and they belong to one of our multiRSEs rule, use that RSE as block location
        :param container: string with the container name
        :param account: string with the account name
        :param scope: string with the scope name (default is "cms")
        :return: a flat dictionary where the keys are the block names, and the value is
          a set with the RSE locations

        NOTE: This is somewhat complex, so I decided to make it more readable with
        a specific method for this process, even though that adds some code duplication.
        """
        result = dict()
        if not self.isContainer(container, scope=scope):
            raise WMRucioException("Pileup location needs to be resolved for a container DID type")

        multiRSERules = []
        finalRSEs = set()
        kargs = dict(name=container, account=account, scope=scope)

        # First, find all the rules and where data is supposed to be locked
        for rule in self.cli.list_replication_rules(kargs):
            rses = self.evaluateRSEExpression(rule['rse_expression'], returnTape=False)
            if rses and rule['copies'] == len(rses) and rule['state'] == "OK":
                # then we can guarantee that data is locked and available on these RSEs
                finalRSEs.update(set(rses))
            # it could be that the rule was made against Tape only, so check
            elif rses:
                multiRSERules.append(rule['id'])
        self.logger.info("Pileup container location for %s from single RSE locks at: %s",
                         kargs['name'], list(finalRSEs))

        # Second, find all the blocks in this pileup container and assign the container
        # level locations to them
        for blockName in self.getBlocksInContainer(kargs['name']):
            result.update({blockName: finalRSEs})
        if not multiRSERules:
            # then that is it, we can return the current RSEs holding and locking this data
            return result

        # if we got here, then there is a third step to be done.
        # List every single block lock and check if the rule belongs to the WMCore system
        for blockName in result:
            blockRSEs = set()
            for blockLock in self.cli.get_dataset_locks(scope, blockName):
                if isTapeRSE(blockLock['rse']):
                    continue
                if blockLock['state'] == 'OK' and blockLock['rule_id'] in multiRSERules:
                    blockRSEs.add(blockLock['rse'])
            result[blockName] = finalRSEs | blockRSEs
        return result

    def getParentContainerRules(self, **kwargs):
        """
        This method takes a DID - such as a file or block - and it resolves its parent
        DID(s). Then it loops over all parent DIDs and - according to the filters
        provided in the kwargs - it lists all their rules.
        :param kwargs: key/value filters supported by list_replication_rules Rucio API, such as:
          * name: string with the DID name (mandatory)
          * scope: string with the scope name (optional)
          * account: string with the rucio account name (optional)
          * state: string with the state name (optional)
          * grouping: string with the grouping name (optional)
          * did_type: string with the DID type (optional)
          * created_before: an RFC-1123 compliant date string (optional)
          * created_after: an RFC-1123 compliant date string (optional)
          * updated_before: an RFC-1123 compliant date string (optional)
          * updated_after: an RFC-1123 compliant date string (optional)
          * and any of the other supported query arguments from the ReplicationRule class, see:
          https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/models.py#L884
        :return: a list of rule ids made against the parent DIDs
        """
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getParentContainerLocation API")
        if 'grouping' in kwargs:
            # long strings seem not to be working, like ALL / DATASET. Make it short!
            kwargs['grouping'] = kwargs['grouping'][0]
        kwargs.setdefault("scope", "cms")
        didName = kwargs['name']

        listOfRules = []
        for parentDID in self.listParentDIDs(kwargs['name']):
            kwargs['name'] = parentDID['name']
            for rule in self.cli.list_replication_rules(kwargs):
                listOfRules.append(rule['id'])
        # revert the original DID name, in case the client will keep using this dict...
        kwargs['name'] = didName
        return listOfRules

    def getDataLockedAndAvailable(self, **kwargs):
        """
        This method retrieves all the locations where a given DID is
        currently available and locked (note that, by default, it will not
        return any Tape RSEs). The logic is as follows:
          1. if DID is a container, then return the result from `getContainerLockedAndAvailable`
          2. resolve the parent DID(s), if any
          3. list all the replication rule ids for the parent DID(s), if any
          4. then lists all the replication rules for this specific DID
          5. then check where blocks are locked and available (state=OK), matching
             one of the replication rule ids discovered in the previous steps
        :param kwargs: key/value filters supported by list_replication_rules Rucio API, such as:
          * name: string with the DID name (mandatory)
          * scope: string with the scope name (optional)
          * account: string with the rucio account name (optional)
          * state: string with the state name (optional)
          * grouping: string with the grouping name (optional)
          * did_type: string with the DID type (optional)
          * created_before: an RFC-1123 compliant date string (optional)
          * created_after: an RFC-1123 compliant date string (optional)
          * updated_before: an RFC-1123 compliant date string (optional)
          * updated_after: an RFC-1123 compliant date string (optional)
          * and any of the other supported query arguments from the ReplicationRule class, see:
          https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/models.py#L884
        :param returnTape: boolean to return Tape RSEs in the output, if any
        :return: a flat list with the RSE names locking and holding the input DID

        NOTE: some of the supported values can be looked up at:
        https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/constants.py#L184
        """
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getBlockLockedAndAvailable API")
        if self.isContainer(kwargs['name'], scope=kwargs.get('scope', 'cms')):
            # then resolve it at container level and all its blocks
            return self.getContainerLockedAndAvailable(**kwargs)

        if 'grouping' in kwargs:
            # long strings seem not to be working, like ALL / DATASET. Make it short!
            kwargs['grouping'] = kwargs['grouping'][0]
        kwargs.setdefault("scope", "cms")
        returnTape = kwargs.pop("returnTape", False)

        finalRSEs = set()
        # first, fetch the rules locking the - possible - parent DIDs
        allRuleIds = self.getParentContainerRules(**kwargs)

        # then lists all the rules for this specific DID
        for rule in self.cli.list_replication_rules(kwargs):
            allRuleIds.append(rule['id'])

        # now with all the rules in hands, we can start checking block locks
        for blockLock in self.cli.get_dataset_locks(kwargs['scope'], kwargs['name']):
            if blockLock['state'] == 'OK' and blockLock['rule_id'] in allRuleIds:
                finalRSEs.add(blockLock['rse'])
        if not returnTape:
            finalRSEs = dropTapeRSEs(finalRSEs)
        else:
            finalRSEs = list(finalRSEs)
        return finalRSEs

    def getContainerLockedAndAvailable(self, **kwargs):
        """
        This method retrieves all the locations where a given container DID is
        currently available and locked (note that, by default, it will not
        return any Tape RSEs). The logic is as follows:
          1. for each container-level replication rule, check
            i. if the rule state=OK and if the number of copies is equals to
             the number of RSEs. If so, that is one of the container locations
            ii. elif the rule state=OK, then consider that rule id (and its RSEs) to
             be evaluated against the dataset locks
            iii. otherwise, don't use the container rule
          2. if there were no rules with multiple RSEs, then return the RSEs from step 1
          3. else, retrieve all the blocks in the container and build a block-based dictionary
            by listing all the dataset_locks for all the blocks:
            i. if the lock state=OK and the rule_id belongs to one of those multiple RSE locks
             from step-1, then use that block location
            ii. else, the location can't be used
          5. finally, build the final container location based on the input `grouping` parameter
           specified, such as:
            i. grouping=ALL triggers an intersection (AND) of all blocks location, which gets added
             to the already discovered container-level location
            ii. other grouping values (like DATASET) triggers an union of all blocks location (note
             that it only takes one block location to consider it as a final location), and a final
             union of this list with the container-level location is made
        :param kwargs: key/value filters supported by list_replication_rules Rucio API, such as:
          * name: string with the DID name (mandatory)
          * scope: string with the scope name (optional)
          * account: string with the rucio account name (optional)
          * state: string with the state name (optional)
          * grouping: string with the grouping name (optional)
          * did_type: string with the DID type (optional)
          * created_before: an RFC-1123 compliant date string (optional)
          * created_after: an RFC-1123 compliant date string (optional)
          * updated_before: an RFC-1123 compliant date string (optional)
          * updated_after: an RFC-1123 compliant date string (optional)
          * and any of the other supported query arguments from the ReplicationRule class, see:
          https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/models.py#L884
        :param returnTape: boolean to return Tape RSEs in the output, if any
        :return: a flat list with the RSE names locking and holding the input DID

        NOTE-1: this is not a full scan of data locking and availability because it does
        not list the replication rules for blocks!!!

        NOTE-2: some of the supported values can be looked up at:
        https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/constants.py#L184
        """
        if 'name' not in kwargs:
            raise WMRucioException("A DID name must be provided to the getContainerLockedAndAvailable API")
        if 'grouping' in kwargs:
            # long strings seem not to be working, like ALL / DATASET. Make it short!
            kwargs['grouping'] = kwargs['grouping'][0]
        kwargs.setdefault("scope", "cms")
        returnTape = kwargs.pop("returnTape", False)

        finalRSEs = set()
        multiRSERules = []
        # first, find all the rules locking this container matching the kwargs
        for rule in self.cli.list_replication_rules(kwargs):
            # now resolve the RSE expressions
            rses = self.evaluateRSEExpression(rule['rse_expression'])
            if rule['copies'] == len(rses) and rule['state'] == "OK":
                # then we can guarantee that data is locked and available on these RSEs
                finalRSEs.update(set(rses))
            elif rule['state'] == "OK":
                if returnTape:
                    multiRSERules.append(rule['id'])
                elif dropTapeRSEs(rses):
                    # if it's not a tape-only rule, use it
                    multiRSERules.append(rule['id'])
            else:
                self.logger.debug("Container rule: %s not yet satisfied. State: %s", rule['id'], rule['state'])
        self.logger.info("Container: %s with container-based location at: %s",
                         kwargs['name'], finalRSEs)
        if not multiRSERules:
            return list(finalRSEs)

        # second, find all the blocks in this container and loop over all of them,
        # checking where they are locked and available
        locationByBlock = dict()
        for block in self.getBlocksInContainer(kwargs['name']):
            locationByBlock.setdefault(block, set())
            for blockLock in self.cli.get_dataset_locks(kwargs['scope'], block):
                if blockLock['state'] == 'OK' and blockLock['rule_id'] in multiRSERules:
                    locationByBlock[block].add(blockLock['rse'])

        # lastly, the final list of RSEs will depend on data grouping requested
        # ALL --> container location is an intersection of each block location
        # DATASET --> container location is the union of each block location.
        #   Note that a block without any location will not affect the final result.
        commonBlockRSEs = set()
        if kwargs.get('grouping') == 'A':
            firstRun = True
            for rses in viewvalues(locationByBlock):
                if firstRun:
                    commonBlockRSEs = set(rses)
                    firstRun = False
                else:
                    commonBlockRSEs = commonBlockRSEs & set(rses)
            # finally, append the block based location to the container rule location
            finalRSEs.update(commonBlockRSEs)
        else:
            for rses in viewvalues(locationByBlock):
                commonBlockRSEs = commonBlockRSEs | set(rses)
            finalRSEs = finalRSEs | commonBlockRSEs

        if not returnTape:
            finalRSEs = dropTapeRSEs(finalRSEs)
        else:
            finalRSEs = list(finalRSEs)
        self.logger.info("Container: %s with block-based location at: %s, and final location: %s",
                         kwargs['name'], commonBlockRSEs, finalRSEs)
        return finalRSEs

    def getRSEUsage(self, rse):
        """
        get_rse_usage rucio API

        :param rse: name of RSE
        :return: generator of records about given RSE
        """
        try:
            return self.cli.get_rse_usage(rse)
        except Exception as exp:
            self.logger.error("Failed to get information about rse %s. Error: %s", rse, str(exp))
            raise exp



"""
File       : FrontPage.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MicroService front page module which defines its web UI.
"""

# futures
from __future__ import print_function, division

# system modules
import os
import re

# WMCore modules
from WMCore.REST.Server import RESTFrontPage


class FrontPage(RESTFrontPage):
    """MicroService front page.
    MicroService provides only one web page, the front page. The page just
    loads the javascript user interface, complete with CSS and all JS
    code embedded into it.

    The JavaScript code performs all the app functionality via the REST
    interface defined by the :class:`~.Data` class.
    """

    def __init__(self, app, config, mount):
        """
        :arg app: reference to the application object.
        :arg config: reference to the configuration.
        :arg str mount: URL mount point."""
        mainroot = 'microservice'  # entry point in access URL
        wpath = os.getenv('MS_STATIC_ROOT', '')
        print(wpath)
        if not wpath:
            content = os.path.abspath(__file__).rsplit('/', 5)[0]
            xlib = (__file__.find("/xlib/") >= 0 and "x") or ""
            wpath = "%s/%sdata/" % (content, xlib)
        if not wpath.endswith('/'):
            wpath += '/'
        print(self.__class__.__name__, "static content: %s" % wpath)
        mdict = {"root": wpath, "rx": re.compile(r"^[a-z]+/[-a-z0-9]+\.(?:html)$")}
        tdict = {"root": wpath + "templates/",
                 "rx": re.compile(r"^([a-zA-Z]+/)*[-a-z0-9_]+\.(?:html|tmpl)$")}
        jdict = {"root": wpath + "js/",
                 "rx": re.compile(r"^([a-zA-Z]+/)*[-a-z0-9_]+\.(?:js)$")}
        cdict = {"root": wpath + "css/",
                 "rx": re.compile(r"^([a-zA-Z]+/)*[-a-z0-9_]+\..*(?:css)$")}
        idict = {"root": wpath + "images/",
                 "rx": re.compile(r"^([a-zA-Z]+/)*[-a-z0-9_]+\.(?:png|gif|jpg)$")}
        roots = {mainroot: mdict, "templates": tdict,
                 "js": jdict, "css": cdict, "images": idict}
        # location of frontpage in the root, e.g. microservice
        frontpage = "%s/templates/index.html" % mainroot
        RESTFrontPage.__init__(self, app, config, mount, frontpage, roots)
